#pragma once

// Opus SDK headers should be included in stdafx.h so their definitions are picked up by both this and LeoHelpers_Opus.cpp.

namespace LeoHelpers
{
#ifdef DOPUS_PLUGIN_HELPER
#ifdef LEOHELPERS_SCOPE_INCLUDED
	// Threadsafe.
	// This loads strings from Directory Opus resources.
	// There's a separate StringLoader that is similar but loads strings using the Win32 ::LoadString.
	class OpusStringLoader
	{
	private:
		mutable LeoHelpers::CritSecWrap m_csw;

		DOpusPluginHelperUtil *m_pHelper;
		// Map is to pointers rather than strings so that it doesn't matter if the map nodes move around in memory.
		std::map< UINT, std::wstring * > m_mapStrings;

		// Disallow copy-constructor and assignment operator:
		OpusStringLoader(const OpusStringLoader &rhs);
		OpusStringLoader &operator=(const OpusStringLoader &rhs);
	public:
		OpusStringLoader(DOpusPluginHelperUtil *pHelper) : m_pHelper(pHelper)
		{
		}

		~OpusStringLoader()
		{
			for (std::map< UINT, std::wstring * >::iterator miter = m_mapStrings.begin(); miter != m_mapStrings.end(); ++miter)
			{
				std::wstring *pwstr = miter->second;
				delete pwstr;
			}
		}

		const wchar_t *Get(UINT id)
		{
			LeoHelpers::CriticalSectionScoper css(m_csw);

			std::wstring *&pwstr = m_mapStrings[id];

			if (pwstr == NULL)
			{
				wchar_t szBuffer[1024];
				wchar_t *pb = szBuffer; // Just to make the compiler happy about __w64 pointer crap.

				SIZE_T res = m_pHelper->GetString(id, &pb, _countof(szBuffer));

				if (res == 0)
				{
					_CrtDbgBreak();
					return L"<missing string>";
				}

				if (res < _countof(szBuffer))
				{
					pwstr = new std::wstring(szBuffer); // this is the map's pointer.
				}
				else
				{
					wchar_t *szLocalAlloc = NULL;
					res = m_pHelper->GetString(id, &szLocalAlloc, 0);

					if (res == 0)
					{
						_CrtDbgBreak();
						return L"<missing string>";
					}

					pwstr = new std::wstring(szLocalAlloc); // this is the map's pointer.
					::LocalFree(szLocalAlloc);
					szLocalAlloc = NULL;

				}
			}

			return pwstr->c_str();
		}

		DOpusPluginHelperUtil *GetOpusPluginHelperUtil() const
		{
			return m_pHelper;
		}
	};

	inline bool XmlGetNodeName(DOpusPluginHelperXML *pHelper, HANDLE hNode, LeoHelpers::AutoBuffer< wchar_t > & scBuffer)
	{
		INT reqSize = 0;

		return (pHelper->XMLGetNodeName(hNode, NULL, &reqSize)
		  &&    scBuffer.AllocateMinimumElements(reqSize)
		  &&    pHelper->XMLGetNodeName(hNode, scBuffer.GetBuffer(), &reqSize));
	}

	inline bool XmlGetNodeValue(DOpusPluginHelperXML *pHelper, HANDLE hNode, LeoHelpers::AutoBuffer< wchar_t > & scBuffer)
	{
		INT reqSize = 0;

		return (pHelper->XMLGetNodeValue(hNode, NULL, &reqSize)
		  &&    scBuffer.AllocateMinimumElements(reqSize)
		  &&    pHelper->XMLGetNodeValue(hNode, scBuffer.GetBuffer(), &reqSize));
	}

	inline bool XmlGetNodeAttribute(DOpusPluginHelperXML *pHelper, HANDLE hNode, const wchar_t *szAttributeName, LeoHelpers::AutoBuffer< wchar_t > & scBuffer)
	{
		INT reqSize = 0;

		return (pHelper->XMLGetNodeAttribute(hNode, szAttributeName, NULL, &reqSize)
		  &&    scBuffer.AllocateMinimumElements(reqSize)
		  &&    pHelper->XMLGetNodeAttribute(hNode, szAttributeName, scBuffer.GetBuffer(), &reqSize));
	}

	// Class used to insert a node if anything actually needs it and avoid creating it if not.
	class XmlOnDemandNode
	{
	private:
		DOpusPluginHelperXML * m_pHelper;
		XmlOnDemandNode *      m_pParentOnDemand; // Allows nested on-demand nodes.
		HANDLE                 m_hParentExplicit;
		HANDLE                 m_hNodeIfCreated;
		std::wstring           m_strNodeName;

		XmlOnDemandNode(const XmlOnDemandNode &rhs); // disallow
		XmlOnDemandNode &operator=(const XmlOnDemandNode &rhs); // disallow

	public:
		XmlOnDemandNode(DOpusPluginHelperXML *pHelper, HANDLE hParentExplicit, const wchar_t *szNodeName)
			: m_pHelper(pHelper)
			, m_pParentOnDemand(NULL)
			, m_hParentExplicit(hParentExplicit)
			, m_hNodeIfCreated(NULL)
			, m_strNodeName(szNodeName)
		{
		}

		XmlOnDemandNode(DOpusPluginHelperXML *pHelper, XmlOnDemandNode *pParentOnDemand, const wchar_t *szNodeName)
			: m_pHelper(pHelper)
			, m_pParentOnDemand(pParentOnDemand)
			, m_hParentExplicit(NULL)
			, m_hNodeIfCreated(NULL)
			, m_strNodeName(szNodeName)
		{
		}

		~XmlOnDemandNode() // Not virtual.
		{
		}

		void Clear()
		{
			m_pParentOnDemand = NULL;
			m_hParentExplicit = NULL;
			m_hNodeIfCreated = NULL;
			m_strNodeName.clear();
		}

		bool IsCreated() const
		{
			return (m_hNodeIfCreated != NULL);
		}

		HANDLE GetNode()
		{
			if (m_hNodeIfCreated != NULL)
			{
				return m_hNodeIfCreated;  // Short-circuit if already created.
			}

			if (m_pParentOnDemand != NULL)
			{
				m_hParentExplicit = m_pParentOnDemand->GetNode();
				m_pParentOnDemand = NULL;
			}

			if (m_hParentExplicit == NULL)
			{
				return NULL;
			}

			m_hNodeIfCreated = m_pHelper->XMLAddChildNode(m_hParentExplicit, m_strNodeName.c_str());

			return m_hNodeIfCreated;
		}
	};

#ifdef LEOHELPERS_FILE_INCLUDED

	inline bool XmlGetConfigPath(DOpusPluginHelperConfig *pHelper, const wchar_t *szConfigBaseName, std::wstring *pStrPath)
	{
		wchar_t szConfigPath[MAX_PATH];

		if (!pHelper->GetConfigPath(OPUSPATH_CONFIG, szConfigPath, _countof(szConfigPath)))
		{
			pStrPath->clear();
			return false;
		}

		*pStrPath = szConfigPath;
		LeoHelpers::AppendPathString(pStrPath, szConfigBaseName);
		pStrPath->append(L".oxc");
		return true;
	}

#endif // LEOHELPERS_FILE_INCLUDED
#endif // LEOHELPERS_SCOPE_INCLUDED
#endif // DOPUS_PLUGIN_HELPER


#ifdef VIEWERPLUGININFO
	inline void WriteFileInfoInfoLine(VIEWERPLUGINFILEINFOW *pInfo, const wchar_t *szTypeName)
	{
		if (NULL != pInfo
		&&	NULL != pInfo->lpszInfo
		&&	0    <  pInfo->cchInfoMax
		&&	NULL != szTypeName)
		{
			if (pInfo->szImageSize.cx > 0
			&&	pInfo->szImageSize.cy > 0
			&&	pInfo->iNumBits > 0)
			{
				_snwprintf_s(pInfo->lpszInfo, pInfo->cchInfoMax, _TRUNCATE, L"%ld x %ld x %d %s", pInfo->szImageSize.cx, pInfo->szImageSize.cy, pInfo->iNumBits, szTypeName);
			}
			else if (pInfo->szImageSize.cx > 0
				 &&	 pInfo->szImageSize.cy > 0)
			{
				_snwprintf_s(pInfo->lpszInfo, pInfo->cchInfoMax, _TRUNCATE, L"%ld x %ld %s", pInfo->szImageSize.cx, pInfo->szImageSize.cy, szTypeName);
			}
			else
			{
				StringCopy(pInfo->lpszInfo, pInfo->cchInfoMax, szTypeName);
			}
		}
	}
#endif

#ifndef DOPUS_PLUGIN_LEO_NO_COPYVERSION

#ifdef VIEWERPLUGININFO
	// If initialisation fails after the call to CopyVersionResourceToViewerPluginInfo then the lpVPInfo->hIconSmall icon must be destroyed.
	bool CopyVersionResourceToViewerPluginInfo(HMODULE hDllModule, WORD iconId, UINT uiOpusMsgIdName, UINT uiOpusMsgIdDesc, LPVIEWERPLUGININFOW lpVPInfo);
#endif

#ifdef VFSPLUGININFO
	// If initialisation fails after the call to CopyVersionResourceToVfsPluginInfo then the lpVPInfo->hIconSmall icon must be destroyed.
	bool CopyVersionResourceToVfsPluginInfo(HMODULE hDllModule, WORD iconId, UINT uiOpusMsgIdName, UINT uiOpusMsgIdDesc, LPVFSPLUGININFO lpVPInfo);
#endif

#endif // DOPUS_PLUGIN_LEO_NO_COPYVERSION
};
